open Core.Std

module Error_code : sig
  type t with sexp

  val is_failure : t -> bool
end

module Context : sig
  type 'a t

  val init : 'a -> 'a t Or_error.t
  val data : 'a t -> 'a
  val error_msg : _ t -> Error_code.t -> string
end

module Keytab : sig
  type t

  val resolve : _ Context.t -> path:string -> t Or_error.t
end

module Principal : sig
  type t

  module Of_sname_type : sig
    type t =
      | NT_SRV_HST
      | NT_UNKNOWN
  end

  val of_sname :
    _ Context.t
    -> service:string
    -> hostname:string
    -> Of_sname_type.t
    -> t Or_error.t

  val of_string :
    _ Context.t
    -> string
    -> t Or_error.t

  val to_string :
    _ Context.t
    -> t
    -> string Or_error.t

  val check_password
    :  _ Context.t
    -> t
    -> password:string
    -> unit Or_error.t

end

module Cred_cache : sig
  type t

  val default : _ Context.t -> t Or_error.t
end

module Peer_context : sig
  type 'a t

  val context : 'a t -> 'a Context.t
end

module Service_context : sig
  type 'a t

  val peer_context : 'a t -> 'a Peer_context.t
  val client_name : _ t -> string

  val init :
    'a Context.t
    -> Principal.t
    -> Keytab.t
    -> in_data:Bigsubstring.t
    -> local_addr:Unix.Inet_addr.t
    -> local_port:int
    -> remote_addr:Unix.Inet_addr.t
    -> remote_port:int
    -> 'a t Or_error.t
end

module Client_context : sig
  type 'a t
  type request = Bigstring.t

  module Flag : sig
    type t =
      | AP_OPTS_USE_SESSION_KEY
      | AP_OPTS_MUTUAL_REQUIRED
  end

  val peer_context : 'a t -> 'a Peer_context.t

  val init :
    ?in_data:Bigsubstring.t
    -> ?replay_id:string
    -> 'a Context.t
    -> Cred_cache.t
    -> Flag.t list
    -> service:string
    -> hostname:string
    -> local_addr:Unix.Inet_addr.t
    -> local_port:int
    -> remote_addr:Unix.Inet_addr.t
    -> remote_port:int
    -> ('a t * request) Or_error.t
end

val mk_priv : _ Peer_context.t -> Bigsubstring.t -> Bigstring.t Or_error.t
val rd_priv : _ Peer_context.t -> Bigsubstring.t -> Bigstring.t Or_error.t

val mk_safe : _ Peer_context.t -> Bigsubstring.t -> Bigstring.t Or_error.t
val rd_safe : _ Peer_context.t -> Bigsubstring.t -> Bigstring.t Or_error.t
