open Core.Std
open Async.Std


val local_inet_of_fd  : Fd.t -> Socket.Address.Inet.t Or_error.t
val remote_inet_of_fd : Fd.t -> Socket.Address.Inet.t Or_error.t

module Error_code : sig
  type t with sexp

  val is_failure : t -> bool
end

module Context : sig
  type 'a t

  val data : 'a t -> 'a
  val init : 'a -> 'a t Deferred.Or_error.t
  val error_msg : _ t -> Error_code.t -> string
end

module Cred_cache : sig
  type t

  val default : _ Context.t -> t Deferred.Or_error.t
end

module Keytab : sig
  type t

  val resolve :
    _ Context.t
    -> path:string
    -> t Deferred.Or_error.t
end

module Principal : sig
  type t

  module Of_sname_type : sig
    type t =
      | NT_SRV_HST
      | NT_UNKNOWN
  end

  val of_sname :
    _ Context.t
    -> service:string
    -> hostname:string
    -> Of_sname_type.t
    -> t Deferred.Or_error.t

  val of_string :
    _ Context.t
    -> string
    -> t Deferred.Or_error.t

  val to_string :
    _ Context.t
    -> t
    -> string Deferred.Or_error.t

  val check_password
    :  _ Context.t
    -> t
    -> password:string
    -> unit Deferred.Or_error.t
end

module Peer_context : sig
  type 'a t

  val context : 'a t -> 'a Context.t
end

module Service_context : sig
  type 'a t

  val peer_context : 'a t -> 'a Peer_context.t
  val client_name : _ t -> string

  val init
    :  'a Context.t
    -> Principal.t
    -> Keytab.t
    -> in_data:Bigsubstring.t
    -> local_inet:Socket.Address.Inet.t
    -> remote_inet:Socket.Address.Inet.t
    -> 'a t Deferred.Or_error.t
(*
  val wrap
    :  Context.t
    -> Principal.t
    -> Keytab.t
    -> local_inet:Socket.Address.Inet.t
    -> remote_inet:Socket.Address.Inet.t
    -> Reader.t -> Writer.t
    -> (t * Reader.t * Writer.t) Deferred.Or_error.t
*)
end

module Client_context : sig
  type 'a t
  type request = Bigstring.t

  module Flag : sig
    type t =
      | AP_OPTS_USE_SESSION_KEY
      | AP_OPTS_MUTUAL_REQUIRED
  end

  val peer_context : 'a t -> 'a Peer_context.t

  val init :
    ?in_data:Bigsubstring.t
    -> ?replay_id:string
    -> 'a Context.t
    -> Cred_cache.t
    -> Flag.t list
    -> service:string
    -> hostname:string
    -> local_inet:Socket.Address.Inet.t
    -> remote_inet:Socket.Address.Inet.t
    -> ('a t * request) Deferred.Or_error.t
end

val mk_priv : _ Peer_context.t -> Bigsubstring.t -> Bigstring.t Or_error.t
val rd_priv : _ Peer_context.t -> Bigsubstring.t -> Bigstring.t Or_error.t

val mk_safe : _ Peer_context.t -> Bigsubstring.t -> Bigstring.t Or_error.t
val rd_safe : _ Peer_context.t -> Bigsubstring.t -> Bigstring.t Or_error.t
